from core.constructs.commands import BaseCommandContainer


class SimpleCommandContainer(BaseCommandContainer):
    help = """
These are the commands that are available for use on the simple static site.

sync: Sync the contents of a given folder to a deployed site

"""
