from core.constructs.commands import BaseCommandContainer


class SimpleCommandContainer(BaseCommandContainer):
    help = """
These are the commands that are available for use on the simple table.

put_items: Load data into the table
clear_table: Clear all data from the table

"""
